@@ -13,9 +13,13 @@ series-specific functionality.
 """
 import struct, threading, time
 from xbee.frame import APIFrame
+from xbee.python2to3 import byteToInt, intToByte
 
 class ThreadQuitException(Exception):
     pass
+    
+class CommandFrameException(KeyError):
+    pass
 
 class XBeeBase(threading.Thread):
     """
@@ -55,7 +59,6 @@ class XBeeBase(threading.Thread):
         if callback:
             self._callback = callback
             self._thread_continue = True
-            self._thread_quit = threading.Event()
             self.start()
 
     def halt(self):
@@ -68,7 +71,7 @@ class XBeeBase(threading.Thread):
         """
         if self._callback:
             self._thread_continue = False
-            self._thread_quit.wait()
+            self.join()
         
     def _write(self, data):
         """
@@ -92,7 +95,6 @@ class XBeeBase(threading.Thread):
                 self._callback(self.wait_read_frame())
             except ThreadQuitException:
                 break
-        self._thread_quit.set()
     
     def _wait_for_frame(self):
         """
@@ -121,14 +123,25 @@ class XBeeBase(threading.Thread):
                 if byte != APIFrame.START_BYTE:
                     continue
 
-                # Save all following bytes
-                frame.fill(byte)
+                # Save all following bytes, if they are not empty
+                if len(byte) == 1:
+                    frame.fill(byte)
+                    
                 while(frame.remaining_bytes() > 0):
-                    frame.fill(self.serial.read())
+                    byte = self.serial.read()
+                    
+                    if len(byte) == 1:
+                        frame.fill(byte)
 
                 try:
                     # Try to parse and return result
                     frame.parse()
+                    
+                    # Ignore empty frames
+                    if len(frame.data) == 0:
+                        frame = APIFrame()
+                        continue
+                        
                     return frame
                 except ValueError:
                     # Bad frame, so restart
@@ -151,7 +164,7 @@ class XBeeBase(threading.Thread):
         except AttributeError:
             raise NotImplementedError("API command specifications could not be found; use a derived class which defines 'api_commands'.")
             
-        packet = ''
+        packet = b''
         
         for field in cmd_spec:
             try:
@@ -202,15 +215,20 @@ class XBeeBase(threading.Thread):
         """
         # Fetch the first byte, identify the packet
         # If the spec doesn't exist, raise exception
-        packet_id = data[0]
+        packet_id = data[0:1]
         try:
             packet = self.api_responses[packet_id]
         except AttributeError:
             raise NotImplementedError("API response specifications could not be found; use a derived class which defines 'api_responses'.")
         except KeyError:
+            # Check to see if this ID can be found among transmittible packets
+            for cmd_name, cmd in list(self.api_commands.items()):
+                if cmd[0]['default'] == data[0:1]:
+                    raise CommandFrameException("Incoming frame with id %s looks like a command frame of type '%s' (these should not be received). Are you sure your devices are in API mode?"
+                            % (data[0], cmd_name))
+            
             raise KeyError(
-                "Unrecognized response packet with id byte %s"
-                % data[0])
+                "Unrecognized response packet with id byte {0}".format(data[0]))
         
         # Current byte index in the data stream
         index = 1
@@ -222,10 +240,10 @@ class XBeeBase(threading.Thread):
         # Parse the packet in the order specified
         for field in packet_spec:
             if field['len'] == 'null_terminated':
-                field_data = ''
+                field_data = b''
                 
-                while data[index] != '\x00':
-                    field_data += data[index]
+                while data[index:index+1] != b'\x00':
+                    field_data += data[index:index+1]
                     index += 1
                 
                 index += 1
@@ -259,13 +277,15 @@ class XBeeBase(threading.Thread):
             raise ValueError(
                 "Response packet was longer than expected; expected: %d, got: %d bytes" % (index, 
                                                                                            len(data)))
-                
-        # Check if this packet was an IO sample
-        # If so, process the sample data
-        if 'parse_as_io_samples' in packet:
-            field_to_process = packet['parse_as_io_samples']
-            info[field_to_process] = self._parse_samples(
-                                        info[field_to_process])
+                                                                                           
+        # Apply parsing rules if any exist
+        if 'parsing' in packet:
+            for parse_rule in packet['parsing']:
+                # Only apply a rule if it is relevant (raw data is available)
+                if parse_rule[0] in info:
+                    # Apply the parse function to the indicated field and 
+                    # replace the raw data with the result
+                    info[parse_rule[0]] = parse_rule[1](self, info)                                                                                   
             
         return info
         
@@ -282,13 +302,13 @@ class XBeeBase(threading.Thread):
         header_size = 3
         
         # number of samples (always 1?) is the first byte
-        sample_count = ord(io_bytes[0])
+        sample_count = byteToInt(io_bytes[0])
         
         # part of byte 1 and byte 2 are the DIO mask ( 9 bits )
-        dio_mask = (ord(io_bytes[1]) << 8 | ord(io_bytes[2])) & 0x01FF
+        dio_mask = (byteToInt(io_bytes[1]) << 8 | byteToInt(io_bytes[2])) & 0x01FF
         
         # upper 7 bits of byte 1 is the AIO mask
-        aio_mask = (ord(io_bytes[1]) & 0xFE) >> 1
+        aio_mask = (byteToInt(io_bytes[1]) & 0xFE) >> 1
         
         # sorted lists of enabled channels; value is position of bit in mask
         dio_chans = []
@@ -326,7 +346,7 @@ class XBeeBase(threading.Thread):
         samples = []
         
         # split the sample data into a list, so it can be pop()'d
-        sample_bytes = [ord(c) for c in io_bytes[header_size:]]
+        sample_bytes = [byteToInt(c) for c in io_bytes[header_size:]]
         
         # repeat for every sample provided
         for sample_ind in range(0, sample_count):
@@ -338,12 +358,11 @@ class XBeeBase(threading.Thread):
                 digital_values = dio_mask & digital_data_set
                 
                 for i in dio_chans:
-                    tmp_samples['dio-%d' % i] = True if (digital_values >> i) & 1 else False
+                    tmp_samples['dio-{0}'.format(i)] = True if (digital_values >> i) & 1 else False
                         
             for i in aio_chans:
-                # only first 10 bits are significant
-                analog_sample = (sample_bytes.pop(0) << 8 | sample_bytes.pop(0)) & 0x03FF
-                tmp_samples['adc-%d' % i] = analog_sample
+                analog_sample = (sample_bytes.pop(0) << 8 | sample_bytes.pop(0))
+                tmp_samples['adc-{0}'.format(i)] = analog_sample
             
             samples.append(tmp_samples)
         
